/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <ctype.h>
#include <sys/stat.h>
#include "jftpgw.h"

extern int multithread;
extern int main_server_pid;
extern int chrooted;
extern struct limitstruct* limitbase;
extern struct hostent_list* hostcache;
extern const int tcp_proto;

int sig_chld_occurred;
int should_read_config;

struct descriptor_set {
	fd_set set;
	int maxfd;
};

static int child_setup(int, struct clientinfo*);
static struct descriptor_set listen_on_ifaces(const char*, struct clientinfo*);
static int get_connecting_socket(struct descriptor_set);
static int say_welcome(int);
static char* prependcode(const char* s, int code);
static void replace_smaller(char* s, char* replace_what, char* replace_with);

/* bindport binds to the specified PORT on HOSTNAME (which may also be a
 * dot-notation IP and returns the socket descriptor
 *
 * Parameters: hostname & port: Where to bind
 *
 * Return value: The socket descriptor of the bound socket
 *
 * Called by: waitclient
 *
 * */

int bindport(const char *hostname, int port) {
	const char* host;
	unsigned long inetaddr =1;
	int shandle;
	int one = 1;
	struct sockaddr_in sin;

	memset((void*)&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	inetaddr = inet_addr(hostname);
	if (inetaddr == (unsigned long int) UINT_MAX) {
		/* HOSTNAME was probably a name since inet_addr returned
		 * an error.
		 * Look up the name to get the IP 
		 */
		host = hostent_get_ip(&hostcache, hostname);
		if (!host) {
			log(1, "Could not resolve %s: %s",
				hostname, strerror(errno));
			perror("Could not resolve the hostname");
			return -1;
		}
		sin.sin_addr.s_addr = inet_addr(gethostentip(host));
	}
	else {
		/* okay, HOSTNAME was a valid dot-notation IP */
		sin.sin_addr.s_addr = inetaddr;
	}
	sin.sin_port = htons(port);

	if (getuid() == 0 && geteuid() != 0) {
		/* become root again - use our function instead of plain
		 * setuid() for the logging */
		log(8, "Changing ID to root (socket(), bind())");
		changeid("root", UID);
	}
	shandle = socket(AF_INET, SOCK_STREAM, tcp_proto);
	if (shandle < 0) {
		log(1, "Error creating socket to bind: %s", strerror(errno));
		perror("Error creating socket to bind to");
		return -1;
	}
	if (setsockopt(shandle, SOL_SOCKET, SO_REUSEADDR,
				(void*) &one, sizeof(one)) < 0) {
		log(3, "Error setting socket to SO_REUSEADDR");
	}
	if (bind(shandle, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		log(1, "Error binding: %s", strerror(errno));
		perror("Error binding");
		return -1;
	}
	if (listen(shandle, 5) < 0) {
		log(1, "Error listening on the socket: %s", strerror(errno));
		perror("Error listening on a bound socket");
		return -1;
	}

	log(6, "Listening on %s, port %d, fd %d", hostname, port, shandle);

	if (getuid() == 0) {
		log(8, "Changing id back (socket(), bind())");
		changeid(config_get_option("runasuser"), EUID);
	}

	return shandle;
}


/* waitclient() waits for a client to connect. It binds to the ports and
 * listens to them. If a connection comes it, jftpgw forks. The parent process
 * keeps on listening whereas the child process handles the connection
 *
 * Parameters: hostnames:      Where to bind to
 *             clntinfo:       connection information
 *
 * Return value: -1 on error (if error message has been sent to the client)
 *               -2 on error (if error message should be created by
 *                            strerror()
 *               socket handle: on success (by the child process
 *
 *               Note: The parent process never returns from this function,
 *                     it is terminated by a signal
 */


int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
	char* peer_ip =0;
	int chldpid =0;
	const char* option = 0;
	int ahandle, ret, i;
	struct sigaction sa;
	struct limitstruct* ls = 0;
	struct descriptor_set d_set;

	if (multithread) {
		daemonize();
	}

	if (getuid() == 0 && geteuid() != 0) {
		/* become root again - use our function instead of plain
		 * setuid() for the logging */
		log(8, "Changing ID to root (pidfile)");
		changeid("root", UID);
	}

	option = config_get_option("pidfile");
	if (option) {
		FILE* pidf = fopen(option, "w");
		if (pidf) {
			fprintf(pidf, "%ld\n", (long) getpid());
			fclose(pidf);
			/* if successful register function to remove the
			 * pidfile */
			atexit(removepidfile);
		} else {
			log(2, "Error creating pidfile %s", option);
		}
	}

	if (getuid() == 0) {
		log(8, "Changing id back (socket(), bind())");
		changeid(config_get_option("runasuser"), EUID);
	}

	/* this has to be done for the daemonization. We do it now after
	 * the pidfile has been created */
	umask(0);

	d_set = listen_on_ifaces(hostnames, clntinfo);
	if (d_set.maxfd < 0) {
		log(8, "d_set.maxfd was negative: %d", d_set.maxfd);
		return -1;
	}

	/* we have successfully bound */

	sa.sa_handler = childterm;
	sig_chld_occurred = 0;
	sigemptyset (&sa.sa_mask);
#ifndef WINDOWS
	sa.sa_flags = SA_RESTART;
#endif
	sigaction (SIGCHLD, &sa, 0);


	/* Close stdin,stdout,stderr */
	for(i = 0; i <= 2 && multithread; i++) {
		close(i);
	}
	main_server_pid = getpid();
	atexit(sayterminating);

	while(1) {
		struct sockaddr_in t_in;
		unsigned long int tg;

		ahandle = get_connecting_socket(d_set);
		if (ahandle == -1) {
			/* either select() or accept() failed */
			/* I don't try resume here because we are in an
			 * endless loop. The danger of the programm falling
			 * into an infinite loop consuming all cpu time is
			 * too big... */
			log(8, "get_connecting_socket() returned error code");
			return -1;
		}

		peer_ip = getpeer_ip(ahandle);
		log(7, "Connection from %s", peer_ip);
		tg = socketinfo_get_local_ip(ahandle);
		t_in.sin_addr.s_addr = tg;
		log(7, "Client tried to connect to %s",
						inet_ntoa(t_in.sin_addr));
		ret = is_below_limit(inet_addr(peer_ip), &ls);
		free(peer_ip);
		if (!ret) {
			say(ahandle, "500 Too many connections from your IP, "
				"sorry\r\n");
			close(ahandle);
			continue;
		}
		if (multithread) {
			if ((chldpid = fork()) < 0) {
				log(1, "Error forking: %s", strerror(errno));
				close(ahandle);
				return -1;
			}
			if (chldpid > 0) {
				/* parent process */
				/* register the PID */
				register_pid(ls, chldpid);
				close(ahandle);
			}
		}
		if (!multithread || chldpid == 0) {
			return child_setup(ahandle, clntinfo);
		}
	}
}


static
int get_connecting_socket(struct descriptor_set d_set) {
#ifdef HAVE_SOCKLEN_T
	socklen_t size;
#else
	int size;
#endif
	static int nfd;
	int shandle, ahandle;
	fd_set backupset;
	struct sockaddr_in sin;

	size = sizeof(sin);

	memcpy(&backupset, &d_set.set, sizeof(fd_set));

	/* is there no remaining ready fd from the last select() ? */
	if (nfd == 0) {
		while (1) {
			/* eternal select() */
			/* nfd returns the number of fds that are ready */
			log(9, "Selecting: d_set.maxfd: %d", d_set.maxfd);
			nfd = select(d_set.maxfd + 1, &d_set.set, 0, 0, 0);
			if (nfd > 0) {
				break;
			}
			if (errno == EINTR) {
				memcpy(&d_set.set, &backupset, sizeof(fd_set));
				if (sig_chld_occurred) {
					get_chld_pid();
				}
				if (should_read_config) {
					log(9, "should read config. Calling function");
					reread_config();
				}
				continue;
			}
			log(1, "select() failed: %s, nfd: %d", strerror(errno), nfd);
			return -1;
		}
	}
	/* a descriptor is ready */
	shandle = 0;
	while (!FD_ISSET(shandle, &d_set.set)) {
		shandle++;
	}
	/* one descriptor less is ready in the set */
	nfd--;

	while(1) {
		ahandle = accept(shandle, (struct sockaddr *) &sin, &size);
		if (ahandle < 0) {
			if (errno == EINTR) {
				continue;
			}
			log(1, "accept() failed: %s", strerror(errno));
			return -1;
		}
		break;
	}
	return ahandle;
}


static
int child_setup(int sock_fd, struct clientinfo *clntinfo) {
	const char* tmp = 0;
	const char* option;
	char* peer_ip = getpeer_ip(sock_fd);
	struct sigaction sa;
	int i, ret;

	/* The clients ignore the SIGHUP signal. Thus
	 * the user can issue a killall -HUP jftpgw
	 * and the master jftpgw process rereads its
	 * configuration file without affecting the
	 * child servers */

	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction(SIGHUP, &sa, 0);

	/* And they just reap the status of exited children
	 */
	sa.sa_handler = reap_chld_info;
	sigemptyset(&sa.sa_mask);
#ifndef WINDOWS
	sa.sa_flags = SA_RESTART;
#endif
	sigaction(SIGCHLD, &sa, 0);

	for (i = 0; i < clntinfo->boundsocket_niface; i++) {
		close(clntinfo->boundsocket_list[i]);
	}
	free(clntinfo->boundsocket_list);
	clntinfo->boundsocket_list = (int*) 0;
	free_limitst(limitbase);
	limitbase = 0;

	log(8, "forked to pid %d", getpid());
	ret = config_shrink_config(inet_addr(peer_ip),	/* source IP */
				-1,		/* dest IP */
				(char*) 0,	/* dest name */
				(char*) 0,	/* dest user */
				(char*) 0,	/* forwarded username */
				0,		/* dest port */
				&hostcache);
	if (ret != 0) {
		log(2, "Error shrinking config data");
		return ret;
	}
	if ( ! config_get_option("access")
		||
		! strcmp(config_get_option("access"), "allow") == 0) {

		say(sock_fd, "500 You are not allowed to "
			"connect. Goodbye.\r\n");
		log(5, "%s was not allowed to connect.",
			peer_ip);
		close(sock_fd);
		free(peer_ip);
		return -2;
	}
	else {
		log(5, "%s is allowed to connect.", peer_ip);
	}

	clntinfo->clientsocket = sock_fd;
	clntinfo->dataport = get_usesport(sock_fd) - 1;
	clntinfo->clientaddr = socketinfo_get_local_ip(sock_fd);

	if (getuid() == 0 && geteuid() != 0) {
		/* become root again to open the logfile(s) in
		 * the logdir(s) */
		log(8, "Changing ID to root (init_logdirs(), "
			"chroot())");
		changeid("root", UID);
	}

	option = config_get_option("changerootdir");
	if (option) {
		if (geteuid() == 0) {
			chdir(option);
			i = chroot(option);
			if (!i) {
				log(7, "Changed root directory "
					"to %s", option);
				chrooted = 1;
			} else {
				log(7, "Error changeing root "
				"directory to %s: %s",
				option, strerror(errno));
				chrooted = 0;
			}
			chdir("/");
		}
	}
	/* chroot()ed environment ? */
	if (!chrooted) {
		tmp = config_get_option("atstartchangerootdir");
		if (!tmp) {
			tmp = config_get_option("changerootdir");
		}
	} else {
		tmp = (char*) 0;
	}
	if (init_logdirs(tmp)) {
		say(sock_fd, "500 Error opening a log file "
			"(see jftpgw log)\r\n");
		free(peer_ip);
		close(sock_fd);
		return -1;
	}

	/* change the effective user id */
	if (getuid() == 0) {
		log(8, "Changing ID back (init_logdirs(), "
			"chroot())");
		if (clntinfo->dataport > IPPORT_RESERVED) {
			/* If we don't need root privileges
			 * any more, change the UID */
			changeid(config_get_option("runasuser"), UID);
		} else {
			changeid(config_get_option("runasuser"), EUID);
		}
	}
	free(peer_ip);
	peer_ip =0;

	say_welcome(sock_fd);

	return sock_fd;
}


static
int say_welcome(int sock_fd) {
	const char* welcome = config_get_option("welcomeline");

	if (config_get_option("welcomeline")) {
		char* welcmsg;

		welcmsg = prependcode(welcome, 220);
		enough_mem(welcmsg);

		log(9, "Saying this text as welcomeline: %s", welcmsg);
		say(sock_fd, welcmsg);
		free(welcmsg);
	} else {
		say(sock_fd, "220 Joe FTP Proxy Server/Gateway (v"JFTPGW_VERSION") ready\r\n");
	}
	return 0;
}


static
struct descriptor_set listen_on_ifaces(const char* hostnames,
				struct clientinfo* clntinfo) {
	char* part =0, *portdel =0;
	char* hostnames2;
	int hostnames2size, offset, port, shandle, i;
	struct descriptor_set d_set;

	offset = 0;
	clntinfo->boundsocket_niface = 0;
	FD_ZERO(&d_set.set);
	d_set.maxfd = 0;

	/* we have to make the string suitable for quotstrtok by appending a
	 * WHITESPACE character */
	hostnames2size = strlen(hostnames) + 2;
	hostnames2 = (char*) malloc(hostnames2size);
	enough_mem(hostnames2);
	snprintf(hostnames2, hostnames2size, "%s ", hostnames);

	while ((part = quotstrtok(hostnames2, WHITESPACES, &offset))) {
		/* do some counting */
		clntinfo->boundsocket_niface++;
	}
	clntinfo->boundsocket_list =
		(int*) malloc(sizeof(int) * clntinfo->boundsocket_niface);
	enough_mem(clntinfo->boundsocket_list);

	offset = 0; i = 0;
	while ((part = quotstrtok(hostnames2, WHITESPACES, &offset))) {
		portdel = strchr(part, ':');
		*portdel = (char) 0;
		errno = 0;
		port = strtol(portdel+1, (char**) 0, 10);
		if (errno) {
			log(4, "Invalid port specification: %s. "
			   "Using default value %d", portdel, DEFAULTBINDPORT);
			port = DEFAULTBINDPORT;
		}
		portdel = (char*) 0;

		log(9, "binding to %s, port %d", part, port);

		shandle = bindport(part, port);
		free(part);
		part = (char*) 0;
		if (shandle < 0) {
			log(8, "Could not bind: %s", strerror(errno));
			free(hostnames2);
			FD_ZERO(&d_set.set);
			d_set.maxfd = -1;
			return d_set;
		}
		FD_SET(shandle, &d_set.set);
		log(9, "Got shandle: %d", shandle);
		d_set.maxfd = MAX_VAL(d_set.maxfd, shandle);
		log(9, "d_set.maxfd is now: %d", d_set.maxfd);
		clntinfo->boundsocket_list[ i++ ] = shandle;
	}
	free(hostnames2);
	hostnames2 = (char*) 0;

	if (clntinfo->boundsocket_niface == 0) {
		log(2, "No interfaces found to bind to");
		FD_ZERO(&d_set.set);
		d_set.maxfd = -1;
		return d_set;
	}

	return d_set;
}


static
char* prependcode(const char* s, int code) {
	char* sdup = strdup(s);
	char* nextline = sdup;
	char* buffer;
	char* end, *replaceend, *bufferend, *bufferstart;
	char dashp = '-';
	const char* const linefeed = "\r\n";
	const char* const empty = "";
	const char* append = linefeed;
	int count = 0;
	int i, length;
	int bufsize;

	enough_mem(sdup);
	replace_smaller(sdup, "\\r", "\r");
	replace_smaller(sdup, "\\n", "\n");

	if (code < 0) {
		code = -code;
	}
	if (code > 999) {
		code = code % 1000;
	}
	if (code < 100) {
		code += 100;
	}

	/* count the newlines */
	while ((nextline = strstr(nextline, "\r\n"))) {
		count++;
		nextline++;  /* thus this position does not match again */
	}

	/* memory: (count + 1) * strlen("xxx ") + 4 (to terminate for sure
	 * with "\r\n") + 1 (Terminator)
	 *
	 * memory = (count + 1) * 4  + 1
	 */

	bufsize = strlen(sdup) + (count + 1) * 4  + 4  + 1;
	bufferstart = buffer = (char*) malloc(bufsize);
	enough_mem(buffer);
	bufferend = buffer + bufsize - 1;

	end = nextline = sdup;
	while (strlen(end)) {
		end = strstr(nextline, "\r\n");
		if (!end) {
			end = nextline + strlen(nextline);
			replaceend = end;
			dashp = ' ';
			append = linefeed;
		} else {
			/* don't replace the first '\' */
			replaceend = end - 1; 
			end += strlen("\r\n");
			if (end == nextline + strlen(nextline)) {
				/* at the very end */
				dashp = ' ';
			} else {
				dashp = '-';
			}
			append = empty;
		}
		i = 0;
		while ((nextline + i) < replaceend) {
			if (nextline[i] == '%') {
				/* Remove the '%' */
				nextline[i] = '/';
			}
			/* remove any special characters between nextline
			 * and end */
			if ((unsigned char) nextline[i] < 32) {
				nextline[i] = '_';
			}
			i++;
		}
		length = MIN_VAL( bufferend - bufferstart,
				  strlen("xxx ")
				  + (end - nextline)
				  + strlen(append)
				  + 1
				);
		snprintf(bufferstart, length, "%d%c%s%s",
				code, dashp, nextline, append);
		bufferstart += strlen(bufferstart);
		if (!*nextline) { /* end */
			nextline = 0;
		}
		nextline = strstr(nextline, "\r\n") + strlen("\r\n");
	}
	free(sdup);
	return buffer;
}


static
void replace_smaller(char* s, char* replace_what, char* replace_with) {
	char* p, *pend;
	p = s;

	while ((p = strstr(p, replace_what))) {
		pend = p + strlen(replace_what);
		snprintf(p, strlen(p), "%s%s", replace_with, pend);
	}
}



