/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <fcntl.h>
#include <ctype.h>
#include <sys/time.h>
#include <arpa/telnet.h> /* for IAC, IP */
#include "jftpgw.h"


/* include the different command sets */

#include "cmds.h"
#include "std_cmds.h"

#ifdef SFTP_SUPPORT
#include "secsftp_cmds.h"
#endif

extern int timeout;
extern int serverport;
extern struct loginstruct lsforward;
extern struct hostent_list* hostcache;
char* portcmd;

struct loginfo li = { NULL, NULL, NULL, 0, 0 };
struct conn_info_st conn_info = { 0, 0, 0, 0 };
struct cmdhandlerstruct *cmdhandler;

int checkforabort(struct clientinfo*);

int handlecmds(const char* hostnames, struct clientinfo *clntinfo) {
	char *buffer = 0;
	char *peer;
	int ss, cs, ret;
	int i;
	struct in_addr iaddr;

	conn_info.lsforward = &lsforward;
	conn_info.li = &li;
	conn_info.clntinfo = clntinfo;

	ss = clntinfo->serversocket = -1;
	clntinfo->cachefd = -1;
	clntinfo->dataclientsock = clntinfo->dataserversock = -1;
	log(8, "setting dataclientsock to -1 (initial)");
	conn_info.usergiven = 0;
	portcmd = 0;
	if ((ret = waitclient(hostnames, clntinfo)) < 0) {
		/* An error occured */
		return ret;
	}
	cs = clntinfo->clientsocket;

	/*
            Server host name                                     svrname
            Server IP address                                    svrip
            Server host name as specified in the login           svrlogin
            Client host name                                     clntname
            Client host IP address                               clntip
            Interface IP to the client                           ifipclnt
            Interface IP to the server                           ifipsvr
	*/


	li.clntip = getpeer_ip(clntinfo->clientsocket);
	li.clntname = hostent_get_name(&hostcache, inet_addr(li.clntip));

	iaddr.s_addr = clntinfo->clientaddr;
	li.ifipclnt = strdup(inet_ntoa(iaddr));
	enough_mem(li.ifipclnt);

	li.filename = (char*) 0;

	while (1) {
contin:
#ifdef SFTP_SUPPORT
		if (clntinfo->servermode == SECURE_SFTP) {
			li.service = "sftp";
			cmdhandler = &secsftp_cmdhandler[0];
		} else {
#endif
			li.service = "ftp";
			cmdhandler = &std_cmdhandler[0];
#ifdef SFTP_SUPPORT
		}
#endif

		ss = clntinfo->serversocket;

		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			return -1;
		}
		if (buffer) {
			/* log the command */
			log_cmd(&li);
			free(buffer);
			buffer = 0;
			li.respcode = 0;
			li.transferred = 0;
		}

		/* See if we're still connected to the client */
		peer = getpeer_ip(cs);
		log(9, "Got %s as client peer", peer);
		if (strcmp(peer, "255.255.255.255") == 0) {
			log(8, "Client closed the connection");
			return -1;
		}
		free(peer);

		if (ss >= 0) {
			/* See if we're still connected to the server */
			peer = getpeer_ip(ss);
			log(9, "Got %s as server peer", peer);
			if (strcmp(peer, "255.255.255.255") == 0) {
				log(8, "Client closed the connection");
				return -1;
			}
			free(peer);
		}

		buffer = readline(cs);
		if (!buffer && timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			err_time_readline(cs);
			return -1;
		}
		if (!buffer) {
			err_readline(cs);
			return -1;
		}

		if (buffer[0] == '\0') {
			/* empty line. Prevent logging of the command */
			free(buffer);
			buffer = 0;
			goto contin;
		}

		li.cmd = buffer;

		free(li.method);
		i = 0;
		li.method = quotstrtok(li.cmd, WHITESPACES, &i);

		/* Retrieve a file */
		clntinfo->mode = RETR;
		i = 0;

		/* If we are not yet connected, the only valid commands are
		 * USER and PASS */
		if (ss < 0 && !checkbegin(buffer, "USER ")
			   && !checkbegin(buffer, "PASS ")
			   && !checkbegin(buffer, "QUIT")) {
			say(cs, "530 Please login with USER and PASS.\r\n");
			li.respcode = 530;
			goto contin;
		}

		while (cmdhandler[i].cmd) {
			if (checkbegin(buffer, cmdhandler[i].cmd)) {
				int ret = (cmdhandler[i].func)
						(buffer, &conn_info);
				if (checkbegin(buffer, "PASS ")) {
					erase_str(buffer);
				}
				switch (ret) {
					case CMD_PASS:
						if (passcmd(buffer, clntinfo)
								< 0) {
							free(buffer);
							return -1;
						}
						break;
					case CMD_HANDLED:
						break;
					case CMD_QUIT:
						free(buffer);
						return 0;
					case CMD_ABORT:
						free(buffer);
						return -1;
				}
				/* found and called proper function */
				goto contin;
			}
			i++;
		}
		/* we didn't find a handler function for the command */
		/* pass it, if we are using the standard command set */

		if (cmdhandler == &std_cmdhandler[0]) {
			if (passcmd(buffer, clntinfo) < 0) {
				free(buffer);
				return -1;
			}
		} else {
			say(clntinfo->clientsocket, 
			"500 Command not understood or not implemented.\r\n");
		}
	}
	/* li.host and li.user  are freed at the termination of the programm */
}



/* getuserdest() parses the "USER xyz" login string */

int getuserdest(const char* cmdline, struct clientinfo *clntinfo) {
	char* orig_login, *portstr, *modestr;
	const char* delim, *opt;
	int i, f, cs, offset, pno;
	char* buffer = strdup(cmdline);

	enough_mem(buffer);

	i = 5;
	f = 0;
	cs = clntinfo->clientsocket;
	clntinfo->servermode = getservermode();

	while (buffer[i] == ' ') {
		i++;
	}
	if(i > strlen(buffer)) {
		say(cs, "500 'USER': command requires a parameter.\r\n");
		li.respcode = 500;
		free(buffer);
		return 1;
	}

	clntinfo->destinationport = 0;

	/* see if we should just pass all the authentication data to another
	 * server. If so we enclose it in "" and add a
	 * "@destinationhost:port" after the login string */

	opt = config_get_option("passallauth");
	if (opt) {
		int bufsize;
		/* passallauth specified */
		/* transform buffer into "buffer"@targethost,21 */

		orig_login = strdup(&buffer[i]);
		enough_mem(orig_login);
		bufsize =  strlen("USER ")
				+ 1                         /* " */
				+ strlen(orig_login)        /* orig_login */
				+ 1                         /* " */
				+ 1                         /* @ */
				+ strlen(opt)               /* opt */
				+ 1;                        /* Termination */
		buffer = (char*) realloc(buffer, bufsize);
		snprintf(buffer, bufsize, "USER \"%s\"@%s", orig_login, opt);
		free(orig_login);
		orig_login = 0;
	}

	/* joe@host,21,p */
	/* joe */
	/* joe@host */
	/* joe,host,21,p */

	delim = "@,:";
	offset = strlen("USER ");

	clntinfo->user = quotstrtok(buffer, delim, &offset);
	clntinfo->destination = quotstrtok(buffer, delim, &offset);
	if (!clntinfo->destination) {
		const char* opt = config_get_option("defaultforward");
		if (opt) {
			/* replace buffer by the forward */
			buffer = realloc(buffer, strlen(opt) + 1);
			strcpy(buffer, opt);
			offset = 0;
			clntinfo->destination = quotstrtok(buffer, delim, &offset);
		}
	}
	if (! clntinfo->destination) {
	/* check for transdest */
		/* opt => dest => forwardhost */
		const char* opt = config_get_option("transparent-forward");
		char* transparent_target;
		offset = 0;

		transparent_target = socketinfo_get_transparent_target(
						clntinfo->clientsocket);
		buffer = realloc(buffer, strlen(transparent_target) + 1);
		strcpy(buffer, transparent_target);
		free(transparent_target); transparent_target = (char*) 0;

		/* Determine the IP the client sees of us */
		clntinfo->clientaddr = socketinfo_get_local_addr_by_sending(
						clntinfo->clientsocket);
		offset = 0;
		clntinfo->destination = quotstrtok(buffer, delim, &offset);

		if (opt) {
			clntinfo->user = realloc(clntinfo->user,
					strlen(clntinfo->user) + 1
					+ strlen(clntinfo->destination) + 1);
			enough_mem(clntinfo->user);
			strcat(clntinfo->user, "@");
			strcat(clntinfo->user, clntinfo->destination);

			/* replace buffer by the forward */
			buffer = realloc(buffer, strlen(opt) + 1);
			strcpy(buffer, opt);
			offset = 0;
			clntinfo->destination = quotstrtok(buffer, delim, &offset);
		}
		log(8, "Using transparent proxy. Connecting to %s. User: %s", clntinfo->destination, clntinfo->user);
	}
	portstr = quotstrtok(buffer, delim, &offset);
	modestr = quotstrtok(buffer, delim, &offset);
	if (portstr) {
		pno = strtol(portstr, NULL, 10);
		if (errno == ERANGE && (pno == LONG_MIN || pno == LONG_MAX)) {
			clntinfo->destinationport = 0;
			free(modestr);
			modestr = portstr;
		} else {
			clntinfo->destinationport = pno;
		}
	} else {
		clntinfo->destinationport = 0;
	}
	if (modestr) {
#ifdef SFTP_SUPPORT
		if (modestr && (strchr("aps", *modestr))) {
#else
		if (modestr && (strchr("ap", *modestr))) {
#endif
			log(9, "mode specified: %s", modestr);
			switch (*modestr) {
				case 'a': 
					clntinfo->servermode = ACTIVE;
					log(9, "p-s: active");
					break;
				case 'p':
					clntinfo->servermode = PASSIVE;
					log(9, "p-s: passive");
					break;
#ifdef SFTP_SUPPORT
				case 's':
					clntinfo->servermode = SECURE_SFTP;
					log(9, "p-s: secure-sftp");
					break;
#endif
			}
		}
	}

	if (!lsforward.fwuser) {
		lsforward.fwuser = extract_username(clntinfo->user);
	}

	li.userlogin     = strdup(lsforward.fwuser);
	li.usereffective = strdup(extract_username(lsforward.fwuser));
	if (lsforward.fwlogin) {
		li.userforwarded = strdup(lsforward.fwlogin);
	} else {
		li.userforwarded = strdup("<no user>");
	}

	if (!clntinfo->destinationport) {
		clntinfo->destinationport = config_get_ioption("serverport",
							DEFAULTSERVERPORT);
	}
	log(7, "Client logged in: User: %s, Dest: %s", clntinfo->user,
							clntinfo->destination);
	switch (clntinfo->servermode) {
		case SECURE_SFTP:
			log(8, "Secure sftp mode");
			break;
		default:
			log(8, "Standard FTP mode, port: %d",
				clntinfo->destinationport);
	}

	/* Get the Password */
	free(buffer);
	return 0;
}



int getpasswd(const char* buffer, struct clientinfo *clntinfo) {
	int i, f, cs;
	char* pass;
	pass = clntinfo->pass = (char*) malloc(strlen(buffer) + 1);
	cs = clntinfo->clientsocket;
	
	f = strlen("PASS ");
	i = 0;
	while (buffer[f] == ' ') {
		f++;
	}
	while (buffer[f]) {
		pass[i] = buffer[f];
		i++; f++;
	}
	pass[i] = '\0';
	while (i > 0 && pass[i] < 32) {
		pass[i] = '\0';
		i--;
	}
	return 0;
}

char* getftpwd(struct clientinfo *clntinfo) {
	char* answer, *dir;
	char* dirstart, *dirend;
	int dirsize;

	say(clntinfo->serversocket, "PWD\r\n");
	answer = ftp_readline(clntinfo->serversocket);
	/*  257 "/home/joe" is current directory.  */
	/*  257 "/home/joe/hi""ho lo di'da" is current directory. */

	if ( ! checkdigits(answer, 257)) {
		log(4, "PWD failed: %s", answer);
		free(answer);
		return (char*) 0;
	}
	dirstart = strchr(answer, '"');
	dirend = strrchr(answer, '"');

	if ( ! dirstart  ||  ! dirend ) {
		log(4, "Could not parse PWD command: %s", answer);
		free(answer);
		return (char*) 0;
	}

	dirstart++;
	dirend--;

	dirsize = dirend - dirstart + 1 + 1;

	dir = (char*) malloc(dirsize);
	enough_mem(dir);

	snprintf(dir, dirsize, "%s", dirstart);
	free(answer);
	char_squeeze(dir, '"');

	return dir;
}

time_t getftpmdtm(const char* filename, struct clientinfo *clntinfo) {
/*
 *	ftp> quote mdtm bla
 *	213 20010224102705
 */
	int cmdsize = strlen("MDTM ") + strlen(filename) + 2 + 1;
	struct tm tms;
	int i;
	char* answer;
	char* cmd = (char*) malloc(cmdsize);

	enough_mem(cmd);
	snprintf(cmd, cmdsize, "MDTM %s\r\n", filename);

	memset(&tms, 0, sizeof(tms));

	say(clntinfo->serversocket, cmd);
	free(cmd);
	answer = ftp_readline(clntinfo->serversocket);
	if ( ! checkdigits(answer, 213)) {
		log(4, "Error reading MDTM answer: %s", answer);
		free(answer);
		return (time_t)(-1);
	}

	/* 213 20010224102705 */
	i = sscanf(answer, "213 %4d%2d%2d%2d%2d%2d",
			&tms.tm_year, &tms.tm_mon, &tms.tm_mday,
			&tms.tm_hour, &tms.tm_min, &tms.tm_sec);

	if (i != 6) {
		log(4, "Error parsing MDTM answer: %s", answer);
		free(answer);
		return (time_t)(-1);
	}
	tms.tm_year -= 1900;  /* tm_year contains the number of years */
			      /* since 1900 */
	tms.tm_mon -= 1;      /* tm_mon starts with 0 */

	free(answer);
	return mktime( &tms );
}

unsigned long int getftpsize(char* filename, struct clientinfo *clntinfo) {
/*
 * ftp> quote size speak.ps
 * 213 146617
 */
	int cmdsize = strlen("SIZE ") + strlen(filename) + 2 + 1;
	unsigned long int size;
	int i;
	char* answer;
	char* cmd = (char*) malloc(cmdsize);

	enough_mem(cmd);
	snprintf(cmd, cmdsize, "SIZE %s\r\n", filename);

	say(clntinfo->serversocket, cmd);
	free(cmd);
	answer = ftp_readline(clntinfo->serversocket);
	if ( ! checkdigits(answer, 213)) {
		log(4, "Error reading SIZE answer: %s", answer);
		free(answer);
		return 0;
	}

	i = sscanf(answer, "213 %ld", &size);
	if (i != 1) {
		log(4, "Error parsing SIZE answer: %s", answer);
		free(answer);
		return 0;
	}

	free(answer);
	return size;
}


int passcmd(const char* buffer, struct clientinfo *clntinfo) {
	int cs = clntinfo->clientsocket;
	int ss = clntinfo->serversocket;
	char* sendbuf =0;
	char *last = 0;
	int ret;
	int sendbufsize;
	int transfer_start;

	sendbufsize = strlen(buffer) + 3;
	sendbuf = (char*) malloc(sendbufsize);
	snprintf(sendbuf, sendbufsize, "%s\r\n", buffer);
	log(9, "Send (%d): %s", ss, sendbuf);
	say(ss, sendbuf);
	free(sendbuf);
	li.complete = 0;
	last = passall(ss, cs);
	if (last) {
		log(9, "Send (%d): %s", cs, last);
	}
	if (!last) {
		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			err_time_readline(cs);
		} else {
			err_readline(cs);
		}
		return -1;
	}
	if (checkdigits(last, 150) || checkdigits(last, 125)) {

		if (portcmd) {
			ret = activeclient(portcmd, clntinfo);
			free(portcmd);
			portcmd = 0;
			if (ret) {
				log(2, "Error in activeclient()");
			}
		}

		if (initiate_transfer(clntinfo) < 0) {
			free(last);
			return -1;
		}

		transfer_start = time(NULL);
		ret = transmitfile(clntinfo);
		li.transfer_duration = time(NULL) - transfer_start;

		if (!ret) {
			char* t = passall(ss, cs);
			li.respcode = respcode(t);
			free(t);
		}
		else {
			if (ret == 1) { /* generate error message */
				if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
					err_time_readline(cs);
				} else {
					err_readline(cs);
				}
			}
			li.respcode = respcode(last);
			free(last);
			return -1;
		}
	}
	/*
	if (checkbegin(buffer, "RETR ") || checkbegin(buffer, "STOR ")) {
		close(clntinfo->dataclientsock);
		close(clntinfo->dataserversock);
		if (clntinfo->cachefd >= 0) {
			close(clntinfo->cachefd);
		}
		log(8, "setting dataclientsock to -1 (reset)");
		clntinfo->dataclientsock = -1;
		clntinfo->dataserversock = -1;
		clntinfo->cachefd        = -1;
	}
	*/
	if (checkbegin(buffer, "RETR ")
			&& clntinfo->tocache
			&& config_get_option("cacheprefix")) {

		if (li.respcode == 226) {
			/* add to cache */
			struct cache_filestruct cfs;
			cfs = cache_gather_info(li.filename, clntinfo);
			cache_add(cfs);
			free(cfs.filepath);
			free(cfs.filename);
		} else {
			/* delete again from cache - should not
			 * happen */
			struct cache_filestruct cfs;
			cfs = cache_gather_info(li.filename, clntinfo);
			cache_delete(cfs);
			free(cfs.filepath);
			free(cfs.filename);
		}
	}
	free(/*at*/ last);  /*  ;-)  */
	last                     =  (char*) 0;
	clntinfo->fromcache      =  0;
	clntinfo->tocache        =  0;
	li.complete = li.respcode == 226;
	return 0;
}


int initiate_transfer(struct clientinfo *clntinfo) {
	int cs = clntinfo->clientsocket;
	int ret;
#ifdef HAVE_SOCKLEN_T
	socklen_t count;
#else
	int count;
#endif
	fd_set acc_set;
	struct timeval tmo = { 300, 0 };
	struct sockaddr_in sin;

	count = sizeof(sin);

	/* if we're talking to the server in active ftp mode, the server
	 * connects to us. Since we're talking to the client in passive
	 * mode, the client connects to us, too! */
	if (clntinfo->servermode == ACTIVE 
	    && clntinfo->clientmode == PASSIVE) {
		/* accept connections from both sides */
		int servacc = 0, clntacc = 0;
		clntinfo->waitforconnect = (int*) 0;
		/* loop until both have connected */
		while (servacc == 0 || clntacc == 0) {
			FD_ZERO(&acc_set);
			if (clntinfo->mode == RETR && clntinfo->fromcache) {
				/* the data comes from the cache */
				close(clntinfo->dataserversock);
				clntinfo->dataserversock = -1;
				servacc = 1;
			} else {
				FD_SET(clntinfo->dataserversock, &acc_set);
			}
			FD_SET(clntinfo->dataclientsock, &acc_set);
			/* select() and wait for both */
			ret = select(MAX_VAL(clntinfo->dataserversock,
					 clntinfo->dataclientsock) + 1,
					&acc_set, NULL, NULL, &tmo);
			if (ret < 0) {
				log(2, "Select() error");
				return -1;
			}
			if (ret == 0) {
				log(2, "Connection timed out");
				li.respcode = 500;
				say(cs, "500 Connection timed out\r\n");
				return -1;
			}
			/* no timeout, no error, one of them must have
			 * connected, check if it was the server. */
			if (FD_ISSET(clntinfo->dataserversock, &acc_set)) {
				ret = accept(clntinfo->dataserversock,
					(struct sockaddr*) &sin, &count);
				if (ret < 0) {
					log(2, "Error in accept() #1: %s",
							strerror(errno));
					return -1;
				}
				close(clntinfo->dataserversock);
				clntinfo->dataserversock = ret;
				servacc = 1;
			}
			/* check if the client connected */
			if (FD_ISSET(clntinfo->dataclientsock, &acc_set)) {
				ret = accept(clntinfo->dataclientsock,
					(struct sockaddr*) &sin, &count);
				if (ret < 0) {
					log(2, "Error in accept() #2: %s",
							strerror(errno));
					return -1;
				}
				close(clntinfo->dataclientsock);
				clntinfo->dataclientsock = ret;
				clntacc = 1;
			}
		}
	}
	if (clntinfo->waitforconnect) {
		if (clntinfo->mode == RETR && clntinfo->fromcache == 1
		&& clntinfo->waitforconnect == &clntinfo->dataserversock) {

			/* the data comes from the cache */
			clntinfo->waitforconnect = (int*) 0;
			close(clntinfo->dataserversock);
			clntinfo->dataserversock = -1;
		} else {
			FD_SET(*clntinfo->waitforconnect, &acc_set);
			ret = select(*clntinfo->waitforconnect + 1, &acc_set,
					NULL, NULL, &tmo);
			if (ret < 0) {
				log(2, "Select() error");
				return -1;
			}
			if (ret == 0) {
				log(2, "Connection timed out");
				li.respcode = 500;
				say(cs, "500 Connection timed out\r\n");
				return -1;
			}

			/* accept connection from one side */
			ret = accept(*clntinfo->waitforconnect,
					(struct sockaddr*) &sin, &count);
			if (ret < 0) {
				log(2, "Error in accept() #3: %s",
						strerror(errno));
				return -1;
			}
			close(*clntinfo->waitforconnect);
			*clntinfo->waitforconnect = ret;
		}
	}

	if (clntinfo->mode == STOR) {
		/* swap descriptors */
		ret = clntinfo->dataserversock;
		clntinfo->dataserversock = clntinfo->dataclientsock;
		clntinfo->dataclientsock = ret;
	}
	return 0;
}

#ifdef WINDOWS
#define TRANSMITBUFSIZE  (10*1024)
#else
#define TRANSMITBUFSIZE  PIPE_BUF
#endif
int transmitfile(struct clientinfo *clntinfo)  {
	char* buffer = (char*) malloc(TRANSMITBUFSIZE);
	char* pbuf = 0;
	int count = 0, nwritten, cachewritten, totwritten, sret = 0, scret = 0;
	int cachefail = 0;
	int cs = clntinfo->clientsocket;
	/* error == 0 : success
	 * error == 1 : fail, generate error message
	 * error == 2 : fail, error message already generated
	 */
	int n, ret, error = 0;
	int fdflags;
	fd_set readset, writeset, exceptset;
	struct timeval readtime = { 300, 0 };
	struct timeval writetime = { 300, 0 }; 
	time_t start, done, delay;
	sigset_t sigset, oldset;

	enough_mem(buffer);
	FD_ZERO(&readset);
	FD_ZERO(&writeset);
	FD_ZERO(&exceptset);
	if (clntinfo->dataserversock == -1 && clntinfo->cachefd >= 0
			&& clntinfo->fromcache && clntinfo->mode == RETR) {
		clntinfo->dataserversock = clntinfo->cachefd;
	}
	FD_SET(clntinfo->dataserversock, &readset);
	FD_SET(cs, &exceptset);
	FD_SET(cs, &readset);
	FD_SET(clntinfo->dataclientsock, &writeset);

	log(7, "Throughputrate is %3.3f", clntinfo->throughput);

	/* set the descriptor nonblocking */
	if ((fdflags = fcntl(clntinfo->dataclientsock, F_GETFL)) < 0) {
		log(2, "Error getting fcntl() flags");
		fdflags = 0;
	}
	ret = fcntl(clntinfo->dataclientsock, F_SETFL, fdflags | O_NONBLOCK);
	if (ret < 0) {
		log(2, "Error setting fd to nonblocking");
	}

	totwritten = 0;
	start = time(NULL);

	sigemptyset(&sigset);
	sigemptyset(&oldset);
	sigaddset(&sigset, SIGCHLD);
	ret = sigprocmask(SIG_BLOCK, &sigset, &oldset);
	if (ret < 0) {
		log(3, "sigprocmask() error: %s", strerror(errno));
	}

	/* read from the clientsocket as well as from the dataserversocket */
	/* no writeset is specified here. The client could send an abort (it
	 * is also in exceptset) and the server sends data */
	while((sret = select(MAX_VAL(clntinfo->dataserversock, cs) + 1,
				&readset, NULL, &exceptset,
				&readtime) > 0)) {
		/* save the errno value of select from sigprocmask() */
		n = errno;
		ret = sigprocmask(SIG_UNBLOCK, &sigset, &oldset);
		if (ret < 0) {
			log(3, "sigprocmask() error releasing the blocked"
				" signals: %s", strerror(errno));
		}
		errno = n;

		/* Can we read data from the client ? */
		if (FD_ISSET(clntinfo->dataserversock, &readset)) {
			count = read(clntinfo->dataserversock,
				buffer, TRANSMITBUFSIZE);
			if (count == 0) {
				log(8, "Read 0 bytes at %s (%d)", __FILE__, __LINE__);
				break;
			}
			if (count < 0) {
				log(3, "read error: %s", strerror(errno));
				/* comm */
				sret = 1;
				error = 1;
				break;
			}
			/* This is only needed for sftp mode connections. If
			 * the proxy acts between a real FTP client and a
			 * real FTP server, both will do the job of
			 * converting */
			if (clntinfo->havetoconvert != CONV_NOTCONVERT) {
				char* tmp;
				if (clntinfo->havetoconvert == CONV_TOASCII) {
					tmp = to_ascii(buffer, &count);
					if (count > TRANSMITBUFSIZE) {
						buffer = realloc(buffer, count);
					}
					memcpy((void*) buffer,
						(void*) tmp, count);
					free(tmp);
				}
				if (clntinfo->havetoconvert == CONV_FRMASCII) {
					tmp = from_ascii(buffer, &count);
					buffer = tmp;
				}
			}

			/* comm */
			pbuf = buffer;
			do {
				/* now write all the read data */
				FD_ZERO(&writeset);
				FD_SET(clntinfo->dataclientsock, &writeset);

				sigemptyset(&sigset);
				sigemptyset(&oldset);
				sigaddset(&sigset, SIGCHLD);
				ret = sigprocmask(SIG_BLOCK, &sigset, &oldset);
				if (ret < 0) {
					log(3, "sigprocmask() error: %s",
						strerror(errno));
				}

				/* selecting to write */
				scret = select(clntinfo->dataclientsock + 1,
					NULL, &writeset, NULL, &writetime);
				/* save the errno value of select from
				 * sigprocmask() */
				n = errno;
				ret = sigprocmask(SIG_UNBLOCK, &sigset,
						&oldset);
				if (ret < 0) {
					log(3, "sigprocmask() error releasing "
						"the blocked signals: %s",
						strerror(errno));
				}
				errno = n;

				if (scret <= 0) {
					break;
				}
				/* otherwise the descriptor must be ready */
				nwritten = write(clntinfo->dataclientsock,
						pbuf, count);
				if (nwritten < 0) {
					log(2, "Error writing (%s, %d): %s",
							__FILE__, __LINE__,
							strerror(errno));
					break;
				}
				if (nwritten == 0) {
					log(8, "Wrote 0 bytes (%s, %d)",
							__FILE__, __LINE__);
				}
				totwritten += nwritten;
				if ( clntinfo->tocache &&  ! cachefail ) {
					cachewritten = write(clntinfo->cachefd,
						pbuf, count);
					if (cachewritten != count) {
						log(3, "Error writing to the "
							"cache: %s",
							strerror(errno));
						cachefail = 1;
					}
				}
				/* calculate the delay time */
				if (clntinfo->throughput >= 0) {
					done = time(NULL);
					delay = (totwritten / 1024.00) /
						clntinfo->throughput;
					delay += start;
					delay -= done;
					if (delay > 0) {
						sleep(delay);
					}
				}
				if (nwritten != count) {
					pbuf += nwritten;
					count -= nwritten;
				} else {
					pbuf = 0;
				}
			} while (pbuf);
		}
		if (FD_ISSET(cs, &exceptset) || FD_ISSET(cs, &readset)) {
			if (checkforabort(clntinfo)) {
				error = 2;
				break;
			}
		}
		/* restore the sets */
		FD_ZERO(&readset);
		FD_ZERO(&writeset);
		FD_ZERO(&exceptset);
		FD_SET(clntinfo->dataserversock, &readset);
		FD_SET(clntinfo->dataclientsock, &writeset);
		FD_SET(cs, &exceptset);
		FD_SET(cs, &readset);
		/* Linux modifies the timeout struct */
		readtime.tv_sec = 300;
		readtime.tv_usec = 0;
		writetime.tv_sec = 300;
		writetime.tv_usec = 0;
	}

	/* Restore flags */
	ret = fcntl(clntinfo->dataclientsock, F_SETFL, fdflags);
	if (ret < 0) {
		log(2, "Error resetting fd flags");
	}

	if (sret == -1 || scret == -1) {
		log(2, "Error in select() occured: %s", strerror(errno));
		li.respcode = 500;
		say(cs, "500 Error in select() occured\r\n");
		error = 2;  /* do not generate an error message */
	}
	/* count == 0 means correct end of transfer */
	if ((sret == 0 || scret == 0) && count != 0) {
		log(2, "Connection timed out in transmitfile(): %d, %d",
				readtime.tv_sec, writetime.tv_sec);
		log(9, "Sockets: dataserver: %d, dataclient: %d, client: %d",
				clntinfo->dataserversock,
				clntinfo->dataclientsock,
				cs);
		if (scret == 0) {
			log(9, "scret == 0 (write)");
		}
		if (sret == 0) {
			log(9, "sret == 0 (read)");
		}
		error = 1;
	}

	close(clntinfo->dataclientsock);
	close(clntinfo->dataserversock);
	if (clntinfo->cachefd >= 0) {
		close(clntinfo->cachefd);
	}
	clntinfo->dataclientsock = -1;
	clntinfo->dataserversock = -1;
	clntinfo->cachefd        = -1;

	clntinfo->fromcache      =  0;

	li.transferred = totwritten;
	log(7, "Transferred %d bytes", totwritten);

	free(buffer);
	clntinfo->havetoconvert = CONV_NOTCONVERT;
	return error;
}

int checkforabort(struct clientinfo* clntinfo) {
	int clientfd = clntinfo->clientsocket;
	int serverfd = clntinfo->serversocket;
	char buffer[1024];
	char* buf =0;
	char rabseq[3] = { (char) IAC, (char) IP, (char) 0 };
	char sabseq[4] = { (char) IAC, (char) IP, (char) IAC, (char) 0 };
	char dmseq[1]  = { (char) DM };
	int ret;

	struct timeval tm = { 300, 0 };
	fd_set set, exset;

	/* this read won't block */
	ret = read(clientfd, buffer, sizeof(buffer) - 1);
	if (ret < 0) {
		/* return, if clientfd is negative for example. "Not a valid
		 * file descriptor". This is important since FD_SET might
		 * segfault with a negative filedescriptor id */
		return 0;
	}
	if (ret == 0) {
		log(8, "Read 0 bytes at %s, %d", __FILE__, __LINE__);
		log(8, "Read from clientfd, closing it! Should not happen!");
		close(clientfd);
		return 0;
	}
	buffer[ret] = '\0';

	if (strncmp(buffer, rabseq, 2) != 0) {
		return 0;
	}

	FD_ZERO(&set);
	FD_SET(clientfd, &set);
	ret = select(clientfd + 1, &set, NULL, &set, &tm);

	if (ret <= 0) {
		return 0;
	}

	ret = read(clientfd, buffer, sizeof(buffer) - 1);
	if (ret < 0) {
		return 0;
	}
	if (ret == 0) {
		log(8, "Read 0 bytes at %s, %d", __FILE__, __LINE__);
		log(8, "Read from clientfd, closing it! Should not happen!");
		close(clientfd);
		return 0;
	}
	buffer[ret] = '\0';

	if(buffer[0] != (char) DM ||
		strncasecmp(buffer + 1,"ABOR",4) != 0) {
		return 0;
	}

	/* abort */

	if (clntinfo->dataclientsock == -1) {
		/* there is no open connection to the client */
		say(clientfd, "225 Abort successful\r\n");
	} else {
		/* Close the connection to the client */
		close(clntinfo->dataclientsock);
		clntinfo->dataclientsock = -1;
		say(clientfd, "426 Transfer aborted. "
				"Closed data connection.\r\n");
		say(clientfd, "226 Abort successful\r\n");
	}

	if (clntinfo->dataserversock == -1) {
		/* if we don't have a connection to the server we can
		 * quit here
		 */
		return 1;
	}
#ifdef SFTP_SUPPORT
	if (clntinfo->servermode == SECURE_SFTP) {
		/* just close the connection */
		cancelled = 1;
		do_cancel();
		close(clntinfo->dataserversock);
		return 1;
	}
#endif

	if (send(serverfd, sabseq, 3, MSG_OOB) != 3
		|| send(serverfd, dmseq, 1, 0) != 1) {
		log(3, "send() error: %s", strerror(errno));
	}
	say(serverfd, "ABOR\r\n");


	FD_ZERO(&set);
	FD_ZERO(&exset);
	FD_SET(serverfd, &set);
	FD_SET(serverfd, &exset);
	tm.tv_sec = 2;  /* grant the server 2 seconds to shut down */
	tm.tv_usec = 0;

	ret = select(serverfd + 1, &set, NULL, &exset, &tm);
	if (ret <= 0) {
		close(clntinfo->dataserversock);
		clntinfo->dataserversock = -1;
	}
	ret = ftp_getrc(serverfd, &buf);
	switch (ret) {
		case 425:
		case 426:
			ret = ftp_getrc(serverfd, 0);
			li.respcode = ret;
			/* fallthrough */
		case 226:
			log(6, "Client requested ABORT - Aborted successfully");
			break;
		default:
			log(6, "Answer of server after ABORT unknown: (%s).",
					buffer);
	}
	free(buf);
	return 1;
}

